#include <io.h>
#include <s3c_gpio.h>
#include <s3c_timer.h>

void udelay(int usec)
{
    volatile unsigned long now, last, diff;

    now = readl(TIMER_BASE + TCNTO4);
    while (usec > 0) {
        last = now;
        now = readl(TIMER_BASE + TCNTO4);
        
        if (now > last)	/* timer has been reloaded */
        {
            diff = last + (TCNTB_MAX - now) + 1;
        }
        else
        {
            diff = last - now;
        }
        usec -= diff;
    }
}

void mdelay(int msec)
{
    udelay(msec * 1000);
}

static int mango_timer_get_tcon_shift_num(int channel)
{
    switch (channel) {
    case 0:	return 0;
    case 1:	return 8;
    case 2:	return 12;
    case 3:	return 16;
    case 4:	return 20;
    default:
        return -1;
    }
}

int mango_timer_configure(int channel, int period, int duty, int invert, int reload)
{
    unsigned long reg;
    unsigned long tcnt, tcmp;
    int shift;

    shift = mango_timer_get_tcon_shift_num(channel);
    if (shift < 0)
        return -1;

#ifdef CONFIG_S3C2443
    if (period > 65535 || duty > 65535)
        return -1;
#endif

    tcnt = 0xC + (0xC * channel);
    tcmp = tcnt + 0x4;
    
    /* update period & duty */
    writel((unsigned long)period, TIMER_BASE + tcnt);
    writel((unsigned long)duty, TIMER_BASE + tcmp);    
    
    reg = readl(TIMER_BASE + TCON);
    reg &= ~(0xf << shift);
    
    if (invert)
        reg |= TCON_TIMER_INVERT << shift;
    if (reload)
        reg |= TCON_TIMER_PERIODIC << shift;

    writel(reg, TIMER_BASE + TCON);

    return 0;
}

int mango_timer_pwm_enable(int channel, S3C_GPIO_PORT port, int pin, int on)
{
    S3C_GPIO_TYPE type;

    if (channel >= MAX_PWM_CHANNEL)
        return -1;

    if (on)
        type = S3C_GPIO_TYPE_ALT1; 
    else
        type = S3C_GPIO_TYPE_OUTPUT;

    return mango_gpio_set_type(port, pin, type);
}

void mango_timer_start(int channel)
{
    unsigned long reg;
    int shift = mango_timer_get_tcon_shift_num(channel);

    reg = readl(TIMER_BASE + TCON);
    writel(reg | (TCON_TIMER_UPDATE << shift), TIMER_BASE + TCON);	/* update TCNTB */
    writel(reg | (TCON_TIMER_START << shift), TIMER_BASE + TCON);	/* start timer */
}

void mango_timer_stop(int channel)
{
    unsigned long reg;
    int shift = mango_timer_get_tcon_shift_num(channel);

    reg = readl(TIMER_BASE + TCON);
    writel(reg & ~(TCON_TIMER_START << shift), TIMER_BASE + TCON);	/* stop timer */
}

void mango_timer_init(void)
{
    unsigned long reg;

    reg = TCFG0_DZLEN(0) | TCFG0_PSCLR1(33) | TCFG0_PSCLR0(1);
    writel(reg, TIMER_BASE + TCFG0);

    reg = TCFG1_DMA_NONE
          | (TCFG1_MUX_1_2 << TCFG1_MUX_SHIFTVAL_TIMER_4)
          | (TCFG1_MUX_1_2 << TCFG1_MUX_SHIFTVAL_TIMER_3)
          | (TCFG1_MUX_1_2 << TCFG1_MUX_SHIFTVAL_TIMER_2)
          | (TCFG1_MUX_1_2 << TCFG1_MUX_SHIFTVAL_TIMER_1)
          | (TCFG1_MUX_1_2 << TCFG1_MUX_SHIFTVAL_TIMER_0);
    writel(reg, TIMER_BASE + TCFG1);

    writel(TCNTB_MAX, TIMER_BASE + TCNTB4);

    reg = TCON_TIMER4_UPDATE;
    writel(reg, TIMER_BASE + TCON);

    reg = TCON_TIMER4_PERIODIC | TCON_TIMER4_START;
    writel(reg, TIMER_BASE + TCON);
}

